local AddonName, Private = ...
local addon = _G[AddonName]

local db
local defaults = {
	sessions = 0,
	someOption = true,
}

local f = CreateFrame("Frame")

local guildBank = {}

local BAGS		= 1		-- All bags, 0 to 11, and keyring ( id -2 )
local BANK		= 2		-- 28 main slots
local GUILDBANK	= 3		-- 98 main slots

local ContainerTypes = {
	[BAGS] = {
		GetSize = function(self, bagID)
				return GetContainerNumSlots(bagID)
			end,
		GetFreeSlots = function(self, bagID)
				-- https://wowpedia.fandom.com/wiki/API_GetContainerNumFreeSlots
				local freeSlots, bagType = GetContainerNumFreeSlots(bagID)
				-- https://wowpedia.fandom.com/wiki/ItemFamily
				return freeSlots, bagType
			end,
		GetLink = function(self, slotID, bagID)
				return GetContainerItemLink(bagID, slotID)
			end,
		GetFrame = function(self, slotID, bagID)
				return _G["ContainerFrame"..bagID.."Item"..slotID]
			end,
		GetCount = function(self, slotID, bagID)
				local _, count = GetContainerItemInfo(bagID, slotID)
				return count
			end,
		GetCooldown = function(self, slotID, bagID)
				local startTime, duration, isEnabled = GetContainerItemCooldown(bagID, slotID)
				return startTime, duration, isEnabled
			end,
	},
	[BANK] = {
		GetSize = function(self)
				return NUM_BANKGENERIC_SLOTS or 28		-- hardcoded in case the constant is not set
			end,
		GetFreeSlots = function(self)
				local freeSlots, bagType = GetContainerNumFreeSlots(-1)		-- -1 = player bank
				return freeSlots, bagType
			end,
		GetLink = function(self, slotID)
				-- return GetInventoryItemLink("player", slotID)
				return GetContainerItemLink(-1, slotID)
			end,
		GetFrame = function(self, slotID, bagID)
				return _G["ContainerFrame"..bagID.."Item"..slotID]
			end,
		GetCount = function(self, slotID)
				-- return GetInventoryItemCount("player", slotID)
				return select(2, GetContainerItemInfo(-1, slotID))
			end,
		GetCooldown = function(self, slotID)
				local startTime, duration, isEnabled = GetInventoryItemCooldown("player", slotID)
				return startTime, duration, isEnabled
			end,
	},
	[GUILDBANK] = {
		GetSize = function(self)
				return MAX_GUILDBANK_SLOTS_PER_TAB or 98		-- hardcoded in case the constant is not set
			end,
		GetFreeSlots = function(self)
				return nil, nil
			end,
		GetLink = function(self, slotID, tabID)
				return GetGuildBankItemLink(tabID, slotID)
			end,
		GetFrame = function(self, slotID, bagID)
				local index = mod(slotID, NUM_SLOTS_PER_GUILDBANK_GROUP);
				if ( index == 0 ) then
					index = NUM_SLOTS_PER_GUILDBANK_GROUP;
				end
				local column = ceil((slotID-0.5)/NUM_SLOTS_PER_GUILDBANK_GROUP);
				return _G["GuildBankColumn"..column.."Button"..index];
			end,
		GetCount = function(self, slotID, tabID)
				local _, count = GetGuildBankItemInfo(tabID, slotID)
				return count
			end,
		GetCooldown = function(self, slotID)
				return nil
			end,
	}
}

-- *** VersionItem Data API ***

function GetIDFromLink(link)
	if link then
		local id = link:match("item:(%d+):");
		if id then 
			return tonumber(id); 
		else 
			return 0; 
		end;
	else
		return -1; 
	end;
end

function getVersion(itemId)
  -- caller must ensure that itemId is a number

  local versionId = Private.itemIdToVersionId[itemId]

  -- item id not found in database
  if not versionId then
    return nil
  end

  return Private.versionIdToVersion[versionId]
end

function getVersionExpac(version)
  return Private.majorToExpac[version.major]
end 

--- *** clean functions ***

local function CleanBagFrame(BagID, slotID, bagframe)
	if bagframe then
		if bagframe:GetParent() and bagframe:GetParent():GetID() then
			if bagframe.ExPack then
				bagframe.ExPack:SetText( "" )
			end
		end
	end
end

-- *** Scanning functions ***

local function ScanBagFrame(BagID, slotID, bagframe, thisContainer)
	if bagframe then
		if bagframe:GetParent() and bagframe:GetParent():GetID() then
			if not bagframe.ExPack then
				bagframe.ExPack = bagframe:CreateFontString(nil, "OVERLAY", "NumberFontNormalSmall")
				bagframe.ExPack:SetPoint("TOPLEFT", 0, 0)
				bagframe.ExPack:SetTextColor(1,1,0) -- yellow
				bagframe.ExPack:SetText("")
				if thisContainer then
					print("ScanBagFrame; "..BagID.."; "..slotID);
				end
			end
			local containerID = bagframe:GetParent():GetID()
			local slot = bagframe:GetID()
			if (contanierID) then
				print("bagframe; contanierID; "..BagID.." <> "..contanierID.."; slotID; "..slotID.." <> "..slot);
			end
			if (slotID ~= slot) then
				--print("bagframe; slotID; "..slotID.." <> "..slot);
			end
			local itemLink
			if thisContainer then
				if (contanierID) then
					if (contanierID ~= BagID) then
						print("thisContainer; contanierID; "..BagID.." <> "..contanierID.."; slotID; "..slotID.." <> "..slot)
					else
						print("thisContainer; contanierID; "..BagID.." = "..contanierID.."; slotID; "..slotID.." <> "..slot)
					end
				else
					print("thisContainer; non-contanierID; "..BagID.."; slotID; "..slotID.." <> "..slot)
					containerID	= BagID -- debug
				end
				itemLink = thisContainer:GetLink(slot, containerID)
			else
				if (contanierID) then
					if (contanierID ~= BagID) then
						print("non-thisContainer; contanierID; "..BagID.." <> "..contanierID"; slotID; "..slotID.." <> "..slot)
					else
						print("non-thisContainer; contanierID; "..BagID.." = "..contanierID.."; slotID; "..slotID.." <> "..slot)
					end
				else
					print("non-thisContainer; non-contanierID; "..BagID.."; slotID; "..slotID.." <> "..slot)
					containerID	= BagID -- debug
				end
				itemLink = GetContainerItemLink(containerID, slot)
			end
			if itemLink then
				-- https://wowpedia.fandom.com/wiki/API_GetItemInfo
				--       1, 2,       3,         4, 5,        6,           7,  8, 9,10,        11,      12, 		   13, 14,      15, 16,                17
				local name, _, quality, itemLevel, _, itemType, itemSubType, _, _,  _, sellPrice, classID, subclassID,  _, expacID,  _, isCraftingReagent = GetItemInfo(itemLink)
				-- mfm
				local expname = "n/a"
				local clr = "|cFFffff00"
				local itemID = GetIDFromLink(itemLink)
				if (itemID > 0) then
					local itemVersion = getVersion(itemID)
					if itemVersion then
						local itemExpac = getVersionExpac(itemVersion)
						expname = itemExpac.shortName
						clr = itemExpac.color
					end
				end

				-- todo
				if bagframe.ExPack then
					if (itemID > 0) then
						local r,g,b
						r, g, b = clr:match("|cFF(%x%x)(%x%x)(%x%x)")
						r, g, b = tonumber(r, 16), tonumber(g, 16), tonumber(b, 16)
						r, g, b = r / 255, g / 255, b / 255
						bagframe.ExPack:SetTextColor(r,g,b)
						bagframe.ExPack:SetText( expname )
					else
						bagframe.ExPack:SetText( "" )
					end
				end
			else
				if bagframe.ExPack then
					bagframe.ExPack:SetText( "" )
				end
			end
		else
			-- print("No bagframe.GetParent; "..BagID..";"..slotID)
			if bagframe.ExPack then
				bagframe.ExPack:SetText( "" )
			end
		end
	else
		--print("No bagframe; "..BagID..";"..slotID)
	end
end

local function ScanContainer(BagID, containerType)
	local thisContainer = ContainerTypes[containerType]
	
	local bag = {}
	bag.size = thisContainer:GetSize(BagID)
	bag.freeslots, bag.bagtype = thisContainer:GetFreeSlots(BagID)

	print("Container; "..containerType.."; "..BagID.."; "..bag.size.."; "..bag.freeslots.."; "..bag.bagtype)

	for slotID=1,bag.size do
		local bagframe = thisContainer:GetFrame(BagID, slotID)
		ScanBagFrame(BagID, slotID, bagframe, thisContainer)
	end
end

local function ScanBag(bagID)
	if bagID < 0 then return end

	ScanContainer(bagID, BAGS)
end

local function ScanReagentBank()
	print("ScanReagentBank")
	ScanContainer(REAGENTBANK_CONTAINER, BANK)
end

-- *** helper functions ***

function GuildBankFrame_RemoveExPack()
	print("GuildBankFrame_RemoveExPack")

	for i=1, MAX_GUILDBANK_SLOTS_PER_TAB do
		index = mod(i, NUM_SLOTS_PER_GUILDBANK_GROUP);
		if ( index == 0 ) then
			index = NUM_SLOTS_PER_GUILDBANK_GROUP;
		end
		column = ceil((i-0.5)/NUM_SLOTS_PER_GUILDBANK_GROUP);
		button = _G["GuildBankColumn"..column.."Button"..index];
		if button then
			CleanBagFrame(page, i, button)
		end
	end
end

-- *** Event Handlers ***

function OnGuildBankFrameOpened()
	print("OnGuildBankFrameOpened")

	guildBank.isGuildBankOpen = true
	guildBank.numTabs = GetNumGuildBankTabs()
	guildBank.guildFunds = GetGuildBankMoney()
	guildBank.tab = GetCurrentGuildBankTab()

--[[
	local tab = GetCurrentGuildBankTab();
	local button, index, column;
	for i=1, MAX_GUILDBANK_SLOTS_PER_TAB do
		index = mod(i, NUM_SLOTS_PER_GUILDBANK_GROUP);
		if ( index == 0 ) then
			index = NUM_SLOTS_PER_GUILDBANK_GROUP;
		end
		column = ceil((i-0.5)/NUM_SLOTS_PER_GUILDBANK_GROUP);
		button = _G["GuildBankColumn"..column.."Button"..index];
		if button then
			ScanBagFrame(tab, i, button)
		else
			if i < 10 then
				print("No GuildBankColumn"..column.."Button"..index)
			end
		end
	end
]]--
	local name, icon
	for page = 1, guildBank.numTabs do
		name, icon = GetGuildBankTabInfo(page)
		
		if not guildBank[page] then
			guildBank[page] = {
				links = {},  --store the slot item links here
				stacks = {}, --store the slot stack counts here
				name = name,
				texture = icon
			}
		end
	end
	noData = false
	waitingForReply = false --if we get our own data from the bank, we don't need to sync
	
	--register the _CLOSED event whenever the bank is opened
	f:RegisterEvent("GUILDBANKFRAME_CLOSED") -- ok
	f:RegisterEvent("GUILDBANKBAGSLOTS_CHANGED")
end

function OnGuildBankFrameChanged()
	print("OnGuildBankFrameChanged")
	local num, itemString
	if (GuildBankFrame and GuildBankFrame:IsVisible()) or (BagnonFrameguildbank and BagnonFrameguildbank:IsVisible()) then

		local page = GetCurrentGuildBankTab()
		if guildBank.tab ~= page then
			--GuildBankFrame_RemoveExPack();
		end
		guildBank.tab = page

		local btnCount = 0
		local thisContainer = ContainerTypes[GUILDBANK]
		for slot=1, MAX_GUILDBANK_SLOTS_PER_TAB do
			index = mod(slot, NUM_SLOTS_PER_GUILDBANK_GROUP);
			if ( index == 0 ) then
				index = NUM_SLOTS_PER_GUILDBANK_GROUP;
			end
			column = ceil((slot-0.5)/NUM_SLOTS_PER_GUILDBANK_GROUP);
			button = _G["GuildBankColumn"..column.."Button"..index];
			if button then
				ScanBagFrame(page, slot, button, thisContainer)
				btnCount = btnCount + 1
			else
				local BankItemLink = GetGuildBankItemLink(page, slot)
			end
		end
		if ( btnCount < MAX_GUILDBANK_SLOTS_PER_TAB ) then
			print("GuildBankColumn; "..btnCount);
		end
		--98 slots on a page
		for slot = 1, 98 do
			--see if the slot has an item in it
			local BankItemLink = GetGuildBankItemLink(page, slot)
			if BankItemLink then
				--if there is link, parse the ID and place that in table w/ key = slot
				_, _, itemString = string.find(GetGuildBankItemLink(page, slot), "^|c%x+|H(.+)|h%[.+%]")
				guildBank[page].links[slot] = itemString
				
				--get item info too, need for stack sizes in image
				_, num = GetGuildBankItemInfo(page, slot)
				guildBank[page].stacks[slot] = num				
			else
				-- that slot is empty
				guildBank[page].links[slot] = nil
				guildBank[page].stacks[slot] = nil
			end
			num, itemString = nil, nil
		end
		
		--save a timestamp for this scan
		guildBank[page].lastScan = time()
	end
end

local function OnGuildBankFrameClosed()
	print("OnGuildBankFrameClosed")
	f:UnregisterEvent("GUILDBANKFRAME_CLOSED") -- ok
	f:UnregisterEvent("GUILDBANKBAGSLOTS_CHANGED")

	-- todo ; claen after close
	--GuildBankFrame_RemoveExPack();
	
	local guildName = GetGuildInfo("player")
	if guildName then
--		GuildBroadcast(MSG_SEND_BANK_TIMESTAMPS, GetBankTimestamps(guildName))
	end
end

--

local function OnBankFrameClosed()
	print("OnBankFrameClosed")
	guildBank.isBankOpen = false
	f:UnregisterEvent("BANKFRAME_CLOSED")
	f:UnregisterEvent("PLAYERBANKSLOTS_CHANGED")
end

local function OnReagentBankFrameChanged()
	print("OnReagentBankFrameChanged")
	if guildBank.isBankOpen then
	end
end

local function OnBankFrameOpened()
	print("OnBankFrameOpened")

	guildBank.isBankOpen = true

	for bagID = NUM_BAG_SLOTS + 1, NUM_BAG_SLOTS + NUM_BANKBAGSLOTS do		-- 5 to 11
		ScanBag(bagID)
	end

	ScanContainer(BANK_CONTAINER, BANK)

	local ReagentBankFrame = _G["ReagentBankFrame"]
	if ReagentBankFrame then
		for slot=1,98 do
			local button = _G["ReagentBankFrameItem"..slot];
			if (button) then
				ScanBagFrame(-3, slot, button)
			end
		end
	else
		print("No ReagentBankFrame available")
		ScanReagentBank()
	end

	--ScanBankSlotsInfo()
	f:RegisterEvent("BANKFRAME_CLOSED", OnBankFrameClosed)
	f:RegisterEvent("PLAYERBANKSLOTS_CHANGED", OnPlayerBankSlotsChanged)
end

---

local function OnEvent(self, event, addOnName)
	-- to-do
	local msg = "OnEvent; before; "..event
	if addOnName then
		msg = msg..";"..addOnName
	end
	print(msg)
	-- Event for guildbank handling
	if ( event == "GUILDBANKFRAME_OPENED" ) then
		OnGuildBankFrameOpened();
		print("OnEvent; after".."; GUILDBANKFRAME_OPENED")
	elseif ( event == "GUILDBANKBAGSLOTS_CHANGED" ) then
		OnGuildBankFrameChanged();
		print("OnEvent; after; "..event)
	elseif ( event == "GUILDBANKFRAME_CLOSED" ) then
		OnGuildBankFrameClosed();
		print("OnEvent; after; "..event)
	elseif ( event == "BANKFRAME_OPENED" ) then
		OnBankFrameOpened();
		print("OnEvent; after".."; "..event)
	elseif ( event == "BANKFRAME_CLOSED" ) then
		print("OnEvent; after".."; "..event)
	elseif ( event == "PLAYERREAGENTBANKSLOTS_CHANGED" ) then
		OnReagentBankFrameChanged();
		print("OnEvent; after".."; "..event)
	end
	-- 
	-- makes a copy instead of pointing Ghordian_ExPackDB to defaults
	-- since tables are passed as a reference
	if addOnName == "Ghordian_ExPack" then
		Ghordian_ExPackDB = Ghordian_ExPackDB or CopyTable(defaults)

		for k, v in pairs(defaults) do
			if Ghordian_ExPackDB[k] == nil then -- don't reset any false values
				Ghordian_ExPackDB[k] = v
			end
		end
		-- use a local variable which points to the DB for readability
		db = Ghordian_ExPackDB
		db.sessions = db.sessions + 1
		print("You loaded this addon "..db.sessions.." times")
		print("someOption is", db.someOption)

		--self:UnregisterEvent(event) -- skip checking for any other addons
	end
end

function ExPack_ScanABag(button)
	--print("ExPack_ScanABag")
	if button then
		local parent = button:GetParent();
		local container = parent:GetID();
		local buttonID = button:GetID();
		ScanBagFrame(container, buttonID, button)
	end
end

function ExPack_ScanBags(event)
	local msg = "ExPack_ScanBags"
	if event then
		msg = msg.."; "..event
	end
	print(msg)
	local MAX_CONTAINER_ITEMS = 36;
	for i=1,NUM_CONTAINER_FRAMES do
		for j=1,MAX_CONTAINER_ITEMS do
			local bagframe = _G["ContainerFrame"..i.."Item"..j]
			ScanBagFrame(i, j, bagframe)
		end
	end
end

-- after the event handler functions

f:RegisterEvent("ADDON_LOADED")

-- Events for Guild bank UI
f:RegisterEvent("GUILDBANKFRAME_OPENED")

-- Events for Player bank UI
f:RegisterEvent("BANKFRAME_OPENED")
f:RegisterEvent("PLAYERREAGENTBANKSLOTS_CHANGED")

f:SetScript("OnEvent", OnEvent)

-- https://wowpedia.fandom.com/wiki/API_OpenAllBags
hooksecurefunc("OpenAllBags",function() ExPack_ScanBags("OpenAllBags") C_Timer.After(0.3, ExPack_ScanBags) end)
hooksecurefunc("ToggleAllBags",function() ExPack_ScanBags("ToggleAllBags") C_Timer.After(0.3, ExPack_ScanBags) end)

-- https://wowpedia.fandom.com/wiki/API_ToggleBag
hooksecurefunc("ToggleBag",function() local bagNum = bagNum or 0; ExPack_ScanBags("ToggleBag;"..bagNum) C_Timer.After(0.3, ExPack_ScanBags) end)

-- https://wowpedia.fandom.com/wiki/API_OpenBag
hooksecurefunc("OpenBag",function() id = id or 0; ExPack_ScanBags("OpenBag;"..id) C_Timer.After(0.3, ExPack_ScanBags) end)

--hooksecurefunc("BankFrameItemButton_Update",function() ExPack_ScanABag() C_Timer.After(0.3, ExPack_ScanABag) end)

-- https://wowpedia.fandom.com/wiki/API_PickupContainerItem
-- PickupContainerItem(bagID, slot)
hooksecurefunc("PickupContainerItem",function() bagID = bagID or 0; slot = slot or 0; ExPack_ScanBags("PickupContainerItem;"..bagID.."; "..slot) C_Timer.After(0.3, ExPack_ScanBags) end)

-- SLASH COMMAND

function ExPack_SlashCommand(msg, editbox)
	-- to-do
	-- enabled/disable overlay on guildbank/reagentbank/inventory/all
	if msg == "reset" then
		Ghordian_ExPackDB = CopyTable(defaults) -- reset to defaults
		db = Ghordian_ExPackDB -- update our table reference
		print("DB has been reset to default")
	elseif msg == "toggle" then
		db.someOption = not db.someOption
		print("Toggled someOption to", db.someOption)
	end
end
 
SLASH_EXPACK1 = "/expack" 
SLASH_EXPACK2 = "/ep"

SlashCmdList.EXPACK = ExPack_SlashCommand;

print("ExPack by Ghordian")
